-- Main.lua

function setup()
    -- Set up the screen
    viewer.mode = FULLSCREEN
    
    -- Initialize variables
    people = {}
    numPeople = 200
    emojiDuration = 1.25 -- seconds
    emojiSize = 180
    emojiFloatSpeed = 15 -- pixels per second
    surpriseProgress = 0
    minSize = math.floor(math.min(WIDTH, HEIGHT) * 0.025)
    maxSize = minSize + 8
    
    -- Personality traits
    local traits = {"shy", "bold", "social"}
    
    for i = 1, numPeople do
        local randomSpeed = minSize * 0.5 * math.random(6, 9)
        local person = {
            x = math.random(WIDTH),
            y = math.random(HEIGHT),
            size = math.random(minSize, maxSize),
            speed = randomSpeed,
            initialSpeed = randomSpeed,
            angle = math.random() * 2 * math.pi,
            emoji = nil,
            emojiTime = 0,
            emojiYOffset = 0,
            color = color(math.random(255), math.random(255), math.random(255)),
            wanderTimer = 0,
            wanderInterval = math.random(10, 80) * 0.01,
            wanderAmount = 3,
            speedChangeTimer = 0,
            speedChangeInterval = math.random(2, 5),
            trait = traits[math.random(#traits)], -- Assign random trait
            speedModifier = 1.0, -- Default speed modifier
            followDistance = math.random(5, 10), -- Default follow distance
            followOffset = math.random(30, 60), -- Default follow offset
            group = nil, -- No leader initially
            groupTimer = 0 -- Timer for switching leaders
        }
        
        -- Adjust properties based on trait
        if person.trait == "shy" then
            person.color = color(0, 0, 255) -- Blue
            person.wanderAmount = 6 -- Shy people change direction more often
        elseif person.trait == "bold" then
            person.color = color(255, 0, 0) -- Red
            person.speedModifier = math.random(25, 30) * 0.1 -- Bold people move faster
            person.wanderAmount = 0.25 -- Bold people change direction less often
        elseif person.trait == "social" then
            person.color = color(0, 255, 0) -- Green
        end
        
        -- Ensure no overlapping spawns
        local overlapping = true
        while overlapping do
            overlapping = false
            for j, other in ipairs(people) do
                if i ~= j and checkCollision(person, other) then
                    person.x = math.random(WIDTH)
                    person.y = math.random(HEIGHT)
                    overlapping = true
                    break
                end
            end
        end
        
        table.insert(people, person)
    end
end

local typeChangeTimer = 0
-- Predefined pastel colors (RGB)
local pastelColors = {
    {255, 209, 220}, -- Pastel pink
    {255, 223, 186}, -- Pastel peach
    {255, 255, 186}, -- Pastel yellow
    {186, 255, 201}, -- Pastel mint
    {186, 225, 255}, -- Pastel blue
    {225, 186, 255}  -- Pastel lavender
}

function draw()
    
    -- Calculate background color
    local colorIndex = (math.sin(ElapsedTime * 0.05) * 0.5 + 0.5) * (#pastelColors - 1) + 1
    local color1 = pastelColors[math.floor(colorIndex)]
    local color2 = pastelColors[math.ceil(colorIndex)]
    local blend = colorIndex - math.floor(colorIndex)
    
    local r = color1[1] * (1 - blend) + color2[1] * blend
    local g = color1[2] * (1 - blend) + color2[2] * blend
    local b = color1[3] * (1 - blend) + color2[3] * blend
    
    background(r, g, b)
    
    -- Variables for text positioning
    local distanceFromTop = HEIGHT * 0.95 -- Distance from the top of the screen
    local lineSpacing = 40 -- Space between lines
    
    -- Draw background text
    pushStyle()
    textMode(CENTER) -- Center-justify text
    fontSize(40) -- Larger text size
    font("AmericanTypewriter-Bold")
    fill(0, 0, 255, 38) -- Faint blue
    text("Blue squares avoid others", WIDTH / 2, distanceFromTop)
    fill(43, 133, 43, 38) -- Faint green
    text("Green squares try to follow others", WIDTH / 2, distanceFromTop - lineSpacing)
    fill(255, 0, 0, 38) -- Faint red
    text("Red squares go fast!", WIDTH / 2, distanceFromTop - 2 * lineSpacing)
    fill(124, 38) -- Faint white
    text("Collisions may change colors", WIDTH / 2, distanceFromTop - 3 * lineSpacing)
    popStyle()
    
    -- Animate surprise lines
    if isTypeChangeActive then
        surpriseProgress = surpriseProgress + DeltaTime / 0.5 -- Animate over 1.5 seconds
        if surpriseProgress > 1 then
            surpriseProgress = 0
            isTypeChangeActive = false -- Reset the flag
        end
    end
    
    -- Draw surprise lines at the collision point if active
    if isTypeChangeActive then
        drawSurpriseLines(collisionX, collisionY, 30, surpriseProgress)
    end
    
    -- Update type change timer
    if isTypeChangeActive then
        typeChangeTimer = typeChangeTimer + DeltaTime
        if typeChangeTimer > 1.5 then -- Reset after 1.5 seconds
            isTypeChangeActive = false
            typeChangeTimer = 0
        end
    end
    
    -- First pass: Draw all people
    for i, person in ipairs(people) do
        
        -- Wander behavior: Randomly change direction
        person.wanderTimer = person.wanderTimer + DeltaTime
        if person.wanderTimer > person.wanderInterval then
            person.angle = person.angle + (math.random() - 0.5) * person.wanderAmount
            person.wanderTimer = 0 -- Reset timer
            person.wanderInterval = math.random(10, 80) * 0.01 -- New random interval
        end
        
        -- update position
        person.x = person.x + math.cos(person.angle) * person.speed * DeltaTime
        person.y = person.y + math.sin(person.angle) * person.speed * DeltaTime
        
        -- Bounce off walls (adjust position and angle)
        if person.x < person.size / 2 then
            person.x = person.size / 2
            person.angle = math.pi - person.angle
        elseif person.x > WIDTH - person.size / 2 then
            person.x = WIDTH - person.size / 2
            person.angle = math.pi - person.angle
        end
        if person.y < person.size / 2 then
            person.y = person.size / 2
            person.angle = -person.angle
        elseif person.y > HEIGHT - person.size / 2 then
            person.y = HEIGHT - person.size / 2
            person.angle = -person.angle
        end
        
        -- Draw the person with their color and rotation
        pushMatrix()
        translate(person.x, person.y) -- Move to the person's position
        rotate(math.deg(person.angle)) -- Convert radians to degrees
        fill(person.color)
        stroke(255)
        strokeWidth(3)
        rect(-person.size / 2, -person.size / 2, person.size, person.size) -- Draw centered
        popMatrix()
        
        -- Occasionally change speed
        person.speedChangeTimer = person.speedChangeTimer + DeltaTime
        if person.speedChangeTimer > person.speedChangeInterval then
            -- Calculate a random speed multiplier relative to the initial speed
            local speedMultiplier = math.random(80, 120) * 0.01 -- Random multiplier between 0.8 and 1.2
            person.speed = person.initialSpeed * speedMultiplier * person.speedModifier -- Apply speed modifier
            person.speedChangeTimer = 0 -- Reset timer
            person.speedChangeInterval = math.random(2, 5) -- New random interval
        end
    end
    
    -- Emoji variables
    local startVerticalOffset = 0
    local endVerticalOffset = 2
    local emojiDuration = 0.9 -- Total time of animation
    local emojiFadeStartTime = 0.65 -- Time to start fade out
    -- Second pass: Draw all emojis
    for i, person in ipairs(people) do
        drawEmoji(person)
    end
    
    -- Check for collisions
    for i = 1, #people do
        for j = i + 1, #people do
            -- Skip if one person is the leader of the other
            if not (people[i].group == people[j] or people[j].group == people[i]) then
                if checkCollision(people[i], people[j]) then
                    if checkCollision(people[i], people[j]) then
                        -- Small chance to trigger a type change
                        if math.random() < 0.05 then
                            -- Decide which person changes type
                            local personToChange = math.random() < 0.5 and people[i] or people[j]
                            changePersonType(personToChange)
                        end
                        
                        -- Handle social grouping
                        if people[i].trait == "social" or people[j].trait == "social" then
                            local socialPerson = people[i].trait == "social" and people[i] or people[j]
                            local targetPerson = people[i].trait == "social" and people[j] or people[i]
                            
                            -- Follow the target person
                            socialPerson.group = targetPerson
                            socialPerson.groupAngle = math.atan(targetPerson.y - socialPerson.y, targetPerson.x - socialPerson.x)
                            socialPerson.groupOffset = math.random(-20, 20)
                            
                            -- Switch targets after some time
                            socialPerson.groupTimer = socialPerson.groupTimer or 0
                            socialPerson.groupTimer = socialPerson.groupTimer + DeltaTime
                            if socialPerson.groupTimer > 5 then -- Switch every 5 seconds
                                socialPerson.group = nil
                                socialPerson.groupTimer = 0
                            end
                        end
                        -- Bounce off each other
                        bounce(people[i], people[j])
                    end
                end
            end
        end
    end
    
    -- Check for social group behavior
    for i, person in ipairs(people) do
        
        if person.trait == "social" then
            -- Find the leader (if any)
            if person.group then
                local leader = person.group
                
                -- Calculate target position behind the leader
                local followDistance = (person.size + leader.size) + person.followDistance -- Adjust multiplier as needed
                local minDistance = 10 -- Minimum distance to maintain
                local dx = leader.x - person.x
                local dy = leader.y - person.y
                local distance = math.sqrt(dx * dx + dy * dy)
                
                if distance > minDistance then
                    local targetX = leader.x - math.cos(leader.angle) * followDistance
                    local targetY = leader.y - math.sin(leader.angle) * followDistance
                    
                    -- Calculate direction to target
                    local targetAngle = math.atan(targetY - person.y, targetX - person.x)
                    
                    -- Smoothly adjust the angle
                    local angleDifference = targetAngle - person.angle
                    if angleDifference > math.pi then
                        angleDifference = angleDifference - 2 * math.pi
                    elseif angleDifference < -math.pi then
                        angleDifference = angleDifference + 2 * math.pi
                    end
                    
                    -- Limit the angle change
                    local maxAngleChange = math.rad(10) -- Maximum angle change per frame (10 degrees)
                    angleDifference = math.max(-maxAngleChange, math.min(maxAngleChange, angleDifference))
                    person.angle = person.angle + angleDifference
                    
                    -- Move toward the target
                    person.x = person.x + math.cos(person.angle) * person.speed * DeltaTime
                    person.y = person.y + math.sin(person.angle) * person.speed * DeltaTime
                end
                
                -- Switch leaders after some time
                person.groupTimer = person.groupTimer or 0
                person.groupTimer = person.groupTimer + DeltaTime
                if person.groupTimer > 5 then -- Switch every 5 seconds
                    person.group = nil
                    person.groupTimer = 0
                end
            else
                -- Find a new leader
                for _, other in ipairs(people) do
                    if other ~= person and (other.trait ~= "social" or other.group ~= person) and checkCollision(person, other, person.size * 3) then
                        person.group = other
                        person.groupTimer = 0
                        break
                    end
                end
            end
        end
        
        -- Shy people avoid collisions
        if person.trait == "shy" then
            for j, other in ipairs(people) do
                if i ~= j and checkCollision(person, other, person.size * 2) then
                    -- Move away from the other person
                    local dx = person.x - other.x
                    local dy = person.y - other.y
                    local distance = math.sqrt(dx * dx + dy * dy)
                    local nx = dx / distance
                    local ny = dy / distance
                    person.angle = math.atan(ny, nx)
                end
            end
        end
    end
end

function drawEmoji(person)
    local startVerticalOffset = 0
    local endVerticalOffset = 2
    local emojiDuration = 0.9 -- Total time of animation
    local emojiFadeStartTime = 0.65 -- Time to start fade out
    
    local startSize = person.size * 0.15
    local endSize = person.size * 2.5
    
    if person.emoji and ElapsedTime - person.emojiTime < emojiDuration then
        local progress = (ElapsedTime - person.emojiTime) / emojiDuration
        
        -- Calculate current size and vertical offset
        local size = startSize + (endSize - startSize) * progress
        local verticalOffset = startVerticalOffset + (endVerticalOffset - startVerticalOffset) * progress
        
        -- Calculate alpha (fade out)
        local alpha = 255
        if progress > emojiFadeStartTime / emojiDuration then
            alpha = 255 * (1 - (progress - emojiFadeStartTime / emojiDuration) / (1 - emojiFadeStartTime / emojiDuration))
        end
        
        -- Draw emoji
        fill(255, 255, 255, alpha)
        fontSize(size)
        text(person.emoji, person.x, person.y + verticalOffset)
    else
        person.emoji = nil
    end
end

function changePersonType(person)
    if isTypeChangeActive then return end -- Skip if a type change is already happening
    isTypeChangeActive = true
    surpriseProgress = 0 -- Start surprise lines animation
    
    -- Calculate collision point (use the person's position)
    collisionX = person.x
    collisionY = person.y
    
    -- Randomly change the person's type, ensuring it's different from their current type
    local traits = {"shy", "bold", "social"}
    local newTrait = person.trait -- Start with the current trait
    while newTrait == person.trait do -- Keep picking until it's different
        newTrait = traits[math.random(#traits)]
    end
    person.trait = newTrait
    person.color = getTraitColor(newTrait)
    
    -- Update properties based on new trait
    if newTrait == "shy" then
        person.wanderAmount = 0.2 -- Shy people change direction more often
        person.speedModifier = 1.0 -- Normal speed
    elseif newTrait == "bold" then
        person.wanderAmount = 2 -- Bold people change direction less often
        person.speedModifier = 2.5 -- Bold people move faster
    elseif newTrait == "social" then
        person.followDistance = math.random(5, 10) -- Random follow distance
        person.followOffset = math.random(30, 60) -- Random follow offset
        person.speedModifier = 1.0 -- Normal speed
    end
    
    -- Show emoji for the person who changed
    person.emoji = getRandomEmoji()
    person.emojiTime = ElapsedTime
    
    -- Play random sound
    playRandomSound()
end

local sounds = {
    asset.downloaded.Game_Sounds_One.Alien_1,
    asset.downloaded.Game_Sounds_One.Alien_2,
    asset.downloaded.Game_Sounds_One.Assembly_1,
    asset.downloaded.Game_Sounds_One.Assembly_2
}

function playRandomSound()
    local soundName = sounds[math.random(#sounds)]
    sound(soundName)
end

function getTraitColor(trait)
    if trait == "shy" then
        return color(0, 0, 255) -- Blue
    elseif trait == "bold" then
        return color(255, 0, 0) -- Red
    elseif trait == "social" then
        return color(0, 255, 0) -- Green
    end
end

function checkCollision(a, b, threshold)
    threshold = threshold or a.size
    local dx = a.x - b.x
    local dy = a.y - b.y
    local distance = math.sqrt(dx * dx + dy * dy)
    return distance < threshold
end

function getRandomEmoji()
    local emojis = {"😀", "😎", "🤖", "👻", "💩", "🎉", "❤️", "🔥", "🌈", "🍕"}
    return emojis[math.random(#emojis)]
end

function bounce(a, b)
    -- Calculate the vector between the two people
    local dx = b.x - a.x
    local dy = b.y - a.y
    local distance = math.sqrt(dx * dx + dy * dy)
    
    -- Normalize the vector
    local nx = dx / distance
    local ny = dy / distance
    
    -- Reflect the angles based on the collision
    a.angle = math.atan(math.sin(a.angle) - ny, math.cos(a.angle) - nx)
    b.angle = math.atan(math.sin(b.angle) + ny, math.cos(b.angle) + nx)
    
    -- Move them apart slightly to avoid sticking
    local overlap = a.size - distance
    a.x = a.x - nx * overlap / 2
    a.y = a.y - ny * overlap / 2
    b.x = b.x + nx * overlap / 2
    b.y = b.y + ny * overlap / 2
end

function drawSurpriseLines(x, y, size, progress)
    local numLines = 25
    local baseLineLength = 30
    local lineWidth = 2
    local angleStep = (math.pi * 2) / numLines
    local startOffset = 2 -- Start 2 pixels away from the circumference
    local endOffset = 25 -- End 17 pixels away from the circumference
    local offset = startOffset + (endOffset - startOffset) * progress
    local alpha = 255
    if progress > 0.75 then
        alpha = 255 * (1 - (progress - 0.95) * 4) -- Faster fade-out
    end
    
    -- Initialize line attributes if not already done
    if not surpriseLineAttributes then
        surpriseLineAttributes = initializeLineAttributes(numLines, baseLineLength)
    end
    
    pushStyle()
    local startPointRadius = (size / 2) + offset
    
    for i = 1, numLines do
        local angle = i * angleStep
        local attribute = surpriseLineAttributes[i]
        local startX = x + startPointRadius * math.cos(angle)
        local startY = y + startPointRadius * math.sin(angle)
        
        -- Line length grows with progress
        local lineLength = attribute.length * progress
        local endX = startX + lineLength * math.cos(angle)
        local endY = startY + lineLength * math.sin(angle)
        
        stroke(attribute.color[1], attribute.color[2], attribute.color[3], alpha)
        strokeWidth(lineWidth)
        line(startX, startY, endX, endY)
    end
    
    if progress >= 1 then
        surpriseLineAttributes = nil -- Reset attributes at the end of the animation
    end
    
    popStyle()
end


function initializeLineAttributes(numLines, baseLineLength)
    local lineAttributes = {}
    for i = 1, numLines do
        -- Ensure lines are never too short
        local minLength = baseLineLength * 0.8 -- Minimum length (80% of base length)
        local maxLength = baseLineLength * 1.2 -- Maximum length (120% of base length)
        local length = math.random(minLength, maxLength) -- Randomize within range
        
        table.insert(lineAttributes, {
            length = length,
            color = {math.random(255), math.random(255), math.random(255)} -- Random color
        })
    end
    return lineAttributes
end

