-- Main.lua
-- Main.lua

function setup()
    initializeScreen()
    initializeVariables()
    createPeople()
end

function draw()
    updateBackground()
    drawInstructions()
    handleSurpriseLines()
    updateAndDrawPeople()
    handleCollisions()
    handleSocialBehavior()
end

-- Helper Functions

function initializeScreen()
    viewer.mode = FULLSCREEN
end

function initializeVariables()
    people = {}
    numPeople = 200
    emojiDuration = 1.25 -- seconds
    surpriseProgress = 0
    minSize = math.floor(math.min(WIDTH, HEIGHT) * 0.025)
    maxSize = math.floor(minSize * 1.2)
    pastelColors = {
        {255, 209, 220}, {255, 223, 186}, {255, 255, 186},
        {186, 255, 201}, {186, 225, 255}, {225, 186, 255}
    }
    typeChangeTimer = 0
end

function createPeople()
    local traits = {"shy", "bold", "social"}
    for i = 1, numPeople do
        local randomSpeed = minSize * 0.5 * math.random(6, 9)
        local person = {
            x = math.random(WIDTH),
            y = math.random(HEIGHT),
            size = math.random(minSize, maxSize),
            speed = randomSpeed,
            initialSpeed = randomSpeed,
            angle = math.random() * 2 * math.pi,
            emoji = nil,
            emojiTime = 0,
            color = color(math.random(255), math.random(255), math.random(255)),
            wanderTimer = 0,
            wanderInterval = math.random(10, 80) * 0.01,
            wanderAmount = 3,
            speedChangeTimer = 0,
            speedChangeInterval = math.random(2, 5),
            trait = traits[math.random(#traits)],
            speedModifier = 1.0,
            followDistance = math.random(5, 10),
            followOffset = math.random(30, 60),
            group = nil,
            groupTimer = 0
        }
        adjustPersonProperties(person)
        ensureNoOverlap(person)
        table.insert(people, person)
    end
end

function adjustPersonProperties(person)
    if person.trait == "shy" then
        person.color = color(0, 0, 255)
        person.wanderAmount = 6
    elseif person.trait == "bold" then
        person.color = color(255, 0, 0)
        person.speedModifier = math.random(25, 30) * 0.1
        person.wanderAmount = 0.25
    elseif person.trait == "social" then
        person.color = color(0, 255, 0)
    end
end

function ensureNoOverlap(person)
    local overlapping = true
    while overlapping do
        overlapping = false
        for _, other in ipairs(people) do
            if checkCollision(person, other) then
                person.x = math.random(WIDTH)
                person.y = math.random(HEIGHT)
                overlapping = true
                break
            end
        end
    end
end

function updateBackground()
    local colorIndex = (math.sin(ElapsedTime * 0.05) * 0.5 + 0.5) * (#pastelColors - 1) + 1
    local color1 = pastelColors[math.floor(colorIndex)]
    local color2 = pastelColors[math.ceil(colorIndex)]
    local blend = colorIndex - math.floor(colorIndex)
    local r = color1[1] * (1 - blend) + color2[1] * blend
    local g = color1[2] * (1 - blend) + color2[2] * blend
    local b = color1[3] * (1 - blend) + color2[3] * blend
    background(r, g, b)
end

function drawInstructions()
    pushStyle()
    textMode(CENTER)
    fontSize(40)
    font("AmericanTypewriter-Bold")
    fill(0, 0, 255, 38)
    text("Blue squares avoid others", WIDTH / 2, HEIGHT * 0.95)
    fill(43, 133, 43, 38)
    text("Green squares try to follow others", WIDTH / 2, HEIGHT * 0.95 - 40)
    fill(255, 0, 0, 38)
    text("Red squares go fast!", WIDTH / 2, HEIGHT * 0.95 - 80)
    fill(124, 38)
    text("Collisions may change colors", WIDTH / 2, HEIGHT * 0.95 - 120)
    popStyle()
end

function handleSurpriseLines()
    if isTypeChangeActive then
        surpriseProgress = surpriseProgress + DeltaTime / 0.5
        if surpriseProgress > 1 then
            surpriseProgress = 0
            isTypeChangeActive = false
        end
        drawSurpriseLines(collisionX, collisionY, 30, surpriseProgress)
    end
end

function updateAndDrawPeople()
    for _, person in ipairs(people) do
        -- Update position and handle wall collisions
        updatePersonPosition(person)
        
        -- Apply shy avoidance behavior (if applicable)
        if person.trait == "shy" then
            handleShyAvoidance(person)
        end
        
        -- Draw the person
        drawPerson(person)
        
        -- Update speed and draw emoji
        updatePersonSpeed(person)
        drawEmoji(person)
    end
end

function updatePersonPosition(person)
    -- Update position based on current angle and speed
    person.x = person.x + math.cos(person.angle) * person.speed * DeltaTime
    person.y = person.y + math.sin(person.angle) * person.speed * DeltaTime
    
    -- Handle wall collisions first
    handleWallBounce(person)
    
    -- Wander behavior
    person.wanderTimer = person.wanderTimer + DeltaTime
    if person.wanderTimer > person.wanderInterval then
        person.angle = person.angle + (math.random() - 0.5) * person.wanderAmount
        person.wanderTimer = 0
        person.wanderInterval = math.random(10, 80) * 0.01
    end
end

function handleWallBounce(person)
    if person.x < person.size / 2 then
        person.x = person.size / 2
        person.angle = math.pi - person.angle -- Bounce off left wall
    elseif person.x > WIDTH - person.size / 2 then
        person.x = WIDTH - person.size / 2
        person.angle = math.pi - person.angle -- Bounce off right wall
    end
    if person.y < person.size / 2 then
        person.y = person.size / 2
        person.angle = -person.angle -- Bounce off bottom wall
    elseif person.y > HEIGHT - person.size / 2 then
        person.y = HEIGHT - person.size / 2
        person.angle = -person.angle -- Bounce off top wall
    end
end

function handleShyAvoidance(person)
    for _, other in ipairs(people) do
        if other ~= person and checkCollision(person, other, person.size * 2) then
            -- Calculate the direction away from the other person
            local dx = person.x - other.x
            local dy = person.y - other.y
            local distance = math.sqrt(dx * dx + dy * dy)
            local nx = dx / distance
            local ny = dy / distance
            
            -- Calculate the avoidance angle
            local avoidanceAngle = math.atan(ny, nx)
            
            -- Adjust the angle smoothly, but don't override wall collision adjustments
            local angleDifference = avoidanceAngle - person.angle
            local maxAngleChange = math.rad(10) -- Maximum angle change per frame (10 degrees)
            angleDifference = math.max(-maxAngleChange, math.min(maxAngleChange, angleDifference))
            person.angle = person.angle + angleDifference
        end
    end
end

function drawPerson(person)
    pushMatrix()
    translate(person.x, person.y)
    rotate(math.deg(person.angle))
    fill(person.color)
    stroke(255)
    strokeWidth(3)
    rect(-person.size / 2, -person.size / 2, person.size, person.size)
    popMatrix()
end

function updatePersonSpeed(person)
    person.speedChangeTimer = person.speedChangeTimer + DeltaTime
    if person.speedChangeTimer > person.speedChangeInterval then
        local speedMultiplier = math.random(40, 80) * 0.01
        person.speed = person.initialSpeed * speedMultiplier * person.speedModifier
        person.speedChangeTimer = 0
        person.speedChangeInterval = math.random(2, 5)
    end
end

function handleCollisions()
    for i = 1, #people do
        for j = i + 1, #people do
            if not (people[i].group == people[j] or people[j].group == people[i]) then
                if checkCollision(people[i], people[j]) then
                    if math.random() < 0.05 then
                        local personToChange = math.random() < 0.5 and people[i] or people[j]
                        changePersonType(personToChange)
                    end
                    bounce(people[i], people[j])
                end
            end
        end
    end
end

function handleSocialBehavior()
    for _, person in ipairs(people) do
        if person.trait == "social" then
            handleSocialFollowing(person)
        elseif person.trait == "shy" then
            handleShyAvoidance(person)
        end
    end
end

function handleSocialFollowing(person)
    if person.group then
        followLeader(person)
    else
        findNewLeader(person)
    end
end

function followLeader(person)
    local leader = person.group
    local followDistance = (person.size + leader.size) + person.followDistance
    local minDistance = 10
    local dx = leader.x - person.x
    local dy = leader.y - person.y
    local distance = math.sqrt(dx * dx + dy * dy)
    if distance > minDistance then
        local targetX = leader.x - math.cos(leader.angle) * followDistance
        local targetY = leader.y - math.sin(leader.angle) * followDistance
        local targetAngle = math.atan(targetY - person.y, targetX - person.x)
        local angleDifference = targetAngle - person.angle
        if angleDifference > math.pi then
            angleDifference = angleDifference - 2 * math.pi
        elseif angleDifference < -math.pi then
            angleDifference = angleDifference + 2 * math.pi
        end
        local maxAngleChange = math.rad(10)
        angleDifference = math.max(-maxAngleChange, math.min(maxAngleChange, angleDifference))
        person.angle = person.angle + angleDifference
        person.x = person.x + math.cos(person.angle) * person.speed * DeltaTime
        person.y = person.y + math.sin(person.angle) * person.speed * DeltaTime
    end
    person.groupTimer = person.groupTimer + DeltaTime
    if person.groupTimer > 5 then
        person.group = nil
        person.groupTimer = 0
    end
end

function findNewLeader(person)
    for _, other in ipairs(people) do
        if other ~= person and (other.trait ~= "social" or other.group ~= person) and checkCollision(person, other, person.size * 3) then
            person.group = other
            person.groupTimer = 0
            break
        end
    end
end

function drawEmoji(person)
    local startVerticalOffset = 0
    local endVerticalOffset = 2
    local emojiDuration = 0.9 -- Total time of animation
    local emojiFadeStartTime = 0.65 -- Time to start fade out
    
    local startSize = person.size * 0.15
    local endSize = person.size * 2.5
    
    if person.emoji and ElapsedTime - person.emojiTime < emojiDuration then
        local progress = (ElapsedTime - person.emojiTime) / emojiDuration
        
        -- Calculate current size and vertical offset
        local size = startSize + (endSize - startSize) * progress
        local verticalOffset = startVerticalOffset + (endVerticalOffset - startVerticalOffset) * progress
        
        -- Calculate alpha (fade out)
        local alpha = 255
        if progress > emojiFadeStartTime / emojiDuration then
            alpha = 255 * (1 - (progress - emojiFadeStartTime / emojiDuration) / (1 - emojiFadeStartTime / emojiDuration))
        end
        
        -- Draw emoji
        fill(255, 255, 255, alpha)
        fontSize(size)
        text(person.emoji, person.x, person.y + verticalOffset)
    else
        person.emoji = nil
    end
end

function changePersonType(person)
    if isTypeChangeActive then return end -- Skip if a type change is already happening
    isTypeChangeActive = true
    surpriseProgress = 0 -- Start surprise lines animation
    
    -- Calculate collision point (use the person's position)
    collisionX = person.x
    collisionY = person.y
    
    -- Randomly change the person's type, ensuring it's different from their current type
    local traits = {"shy", "bold", "social"}
    local newTrait = person.trait -- Start with the current trait
    while newTrait == person.trait do -- Keep picking until it's different
        newTrait = traits[math.random(#traits)]
    end
    person.trait = newTrait
    person.color = getTraitColor(newTrait)
    
    -- Update properties based on new trait
    if newTrait == "shy" then
        person.wanderAmount = 0.2 -- Shy people change direction more often
        person.speedModifier = 1.0 -- Normal speed
    elseif newTrait == "bold" then
        person.wanderAmount = 2 -- Bold people change direction less often
        person.speedModifier = 2.5 -- Bold people move faster
    elseif newTrait == "social" then
        person.followDistance = math.random(5, 10) -- Random follow distance
        person.followOffset = math.random(30, 60) -- Random follow offset
        person.speedModifier = 1.0 -- Normal speed
    end
    
    -- Show emoji for the person who changed
    person.emoji = getRandomEmoji()
    person.emojiTime = ElapsedTime
    
    -- Play random sound
    playRandomSound()
end

local sounds = {
    asset.downloaded.Game_Sounds_One.Alien_1,
    asset.downloaded.Game_Sounds_One.Alien_2,
    asset.downloaded.Game_Sounds_One.Assembly_1,
    asset.downloaded.Game_Sounds_One.Assembly_2
}

function playRandomSound()
    local soundName = sounds[math.random(#sounds)]
    sound(soundName)
end

function getTraitColor(trait)
    if trait == "shy" then
        return color(0, 0, 255) -- Blue
    elseif trait == "bold" then
        return color(255, 0, 0) -- Red
    elseif trait == "social" then
        return color(0, 255, 0) -- Green
    end
end

function checkCollision(a, b, threshold)
    threshold = threshold or a.size
    local dx = a.x - b.x
    local dy = a.y - b.y
    local distance = math.sqrt(dx * dx + dy * dy)
    return distance < threshold
end

function getRandomEmoji()
    local emojis = {"😀", "😎", "🤖", "👻", "💩", "🎉", "❤️", "🔥", "🌈", "🍕"}
    return emojis[math.random(#emojis)]
end

function bounce(a, b)
    -- Calculate the vector between the two people
    local dx = b.x - a.x
    local dy = b.y - a.y
    local distance = math.sqrt(dx * dx + dy * dy)
    
    -- Normalize the vector
    local nx = dx / distance
    local ny = dy / distance
    
    -- Reflect the angles based on the collision
    a.angle = math.atan(math.sin(a.angle) - ny, math.cos(a.angle) - nx)
    b.angle = math.atan(math.sin(b.angle) + ny, math.cos(b.angle) + nx)
    
    -- Move them apart slightly to avoid sticking
    local overlap = a.size - distance
    a.x = a.x - nx * overlap / 2
    a.y = a.y - ny * overlap / 2
    b.x = b.x + nx * overlap / 2
    b.y = b.y + ny * overlap / 2
end

function drawSurpriseLines(x, y, size, progress)
    local numLines = 25
    local baseLineLength = 30
    local lineWidth = 2
    local angleStep = (math.pi * 2) / numLines
    local startOffset = 2 -- Start 2 pixels away from the circumference
    local endOffset = 25 -- End 17 pixels away from the circumference
    local offset = startOffset + (endOffset - startOffset) * progress
    local alpha = 255
    if progress > 0.75 then
        alpha = 255 * (1 - (progress - 0.95) * 4) -- Faster fade-out
    end
    
    -- Initialize line attributes if not already done
    if not surpriseLineAttributes then
        surpriseLineAttributes = initializeLineAttributes(numLines, baseLineLength)
    end
    
    pushStyle()
    local startPointRadius = (size / 2) + offset
    
    for i = 1, numLines do
        local angle = i * angleStep
        local attribute = surpriseLineAttributes[i]
        local startX = x + startPointRadius * math.cos(angle)
        local startY = y + startPointRadius * math.sin(angle)
        
        -- Line length grows with progress
        local lineLength = attribute.length * progress
        local endX = startX + lineLength * math.cos(angle)
        local endY = startY + lineLength * math.sin(angle)
        
        stroke(attribute.color[1], attribute.color[2], attribute.color[3], alpha)
        strokeWidth(lineWidth)
        line(startX, startY, endX, endY)
    end
    
    if progress >= 1 then
        surpriseLineAttributes = nil -- Reset attributes at the end of the animation
    end
    
    popStyle()
end


function initializeLineAttributes(numLines, baseLineLength)
    local lineAttributes = {}
    for i = 1, numLines do
        -- Ensure lines are never too short
        local minLength = baseLineLength * 0.8 -- Minimum length (80% of base length)
        local maxLength = baseLineLength * 1.2 -- Maximum length (120% of base length)
        local length = math.random(minLength, maxLength) -- Randomize within range
        
        table.insert(lineAttributes, {
            length = length,
            color = {math.random(255), math.random(255), math.random(255)} -- Random color
        })
    end
    return lineAttributes
end

function draw()
    -- Local functions for modularity
    local function updateBackground()
        local colorIndex = (math.sin(ElapsedTime * 0.05) * 0.5 + 0.5) * (#pastelColors - 1) + 1
        local color1 = pastelColors[math.floor(colorIndex)]
        local color2 = pastelColors[math.ceil(colorIndex)]
        local blend = colorIndex - math.floor(colorIndex)
        local r = color1[1] * (1 - blend) + color2[1] * blend
        local g = color1[2] * (1 - blend) + color2[2] * blend
        local b = color1[3] * (1 - blend) + color2[3] * blend
        background(r, g, b)
    end
    
    local function drawInstructions()
        local distanceFromTop = HEIGHT * 0.95
        local lineSpacing = 40
        pushStyle()
        textMode(CENTER)
        fontSize(40)
        font("AmericanTypewriter-Bold")
        fill(0, 0, 255, 38)
        text("Blue squares avoid others", WIDTH / 2, distanceFromTop)
        fill(43, 133, 43, 38)
        text("Green squares try to follow others", WIDTH / 2, distanceFromTop - lineSpacing)
        fill(255, 0, 0, 38)
        text("Red squares go fast!", WIDTH / 2, distanceFromTop - 2 * lineSpacing)
        fill(124, 38)
        text("Collisions may change colors", WIDTH / 2, distanceFromTop - 3 * lineSpacing)
        popStyle()
    end
    
    local function handleSurpriseLines()
        if isTypeChangeActive then
            surpriseProgress = surpriseProgress + DeltaTime / 0.5
            if surpriseProgress > 1 then
                surpriseProgress = 0
                isTypeChangeActive = false
            end
            drawSurpriseLines(collisionX, collisionY, 30, surpriseProgress)
        end
    end
    
    local function updateTypeChangeTimer()
        if isTypeChangeActive then
            typeChangeTimer = typeChangeTimer + DeltaTime
            if typeChangeTimer > 1.5 then
                isTypeChangeActive = false
                typeChangeTimer = 0
            end
        end
    end
    
    local function updateAndDrawPeople()
        for i, person in ipairs(people) do
            -- Wander behavior
            person.wanderTimer = person.wanderTimer + DeltaTime
            if person.wanderTimer > person.wanderInterval then
                person.angle = person.angle + (math.random() - 0.5) * person.wanderAmount
                person.wanderTimer = 0
                person.wanderInterval = math.random(10, 80) * 0.01
            end
            
            -- Update position
            person.x = person.x + math.cos(person.angle) * person.speed * DeltaTime
            person.y = person.y + math.sin(person.angle) * person.speed * DeltaTime
            
            -- Bounce off walls
            if person.x < person.size / 2 then
                person.x = person.size / 2
                person.angle = math.pi - person.angle
            elseif person.x > WIDTH - person.size / 2 then
                person.x = WIDTH - person.size / 2
                person.angle = math.pi - person.angle
            end
            if person.y < person.size / 2 then
                person.y = person.size / 2
                person.angle = -person.angle
            elseif person.y > HEIGHT - person.size / 2 then
                person.y = HEIGHT - person.size / 2
                person.angle = -person.angle
            end
            
            -- Draw the person
            pushMatrix()
            translate(person.x, person.y)
            rotate(math.deg(person.angle))
            fill(person.color)
            stroke(255)
            strokeWidth(3)
            rect(-person.size / 2, -person.size / 2, person.size, person.size)
            popMatrix()
            
            -- Occasionally change speed
            person.speedChangeTimer = person.speedChangeTimer + DeltaTime
            if person.speedChangeTimer > person.speedChangeInterval then
                local speedMultiplier = math.random(80, 120) * 0.01
                person.speed = person.initialSpeed * speedMultiplier * person.speedModifier
                person.speedChangeTimer = 0
                person.speedChangeInterval = math.random(2, 5)
            end
        end
    end
    
    local function drawEmojis()
        for i, person in ipairs(people) do
            drawEmoji(person)
        end
    end
    
    local function handleCollisions()
        for i = 1, #people do
            for j = i + 1, #people do
                if not (people[i].group == people[j] or people[j].group == people[i]) then
                    if checkCollision(people[i], people[j]) then
                        if math.random() < 0.05 then
                            local personToChange = math.random() < 0.5 and people[i] or people[j]
                            changePersonType(personToChange)
                        end
                        
                        -- Handle social grouping
                        if people[i].trait == "social" or people[j].trait == "social" then
                            local socialPerson = people[i].trait == "social" and people[i] or people[j]
                            local targetPerson = people[i].trait == "social" and people[j] or people[i]
                            
                            socialPerson.group = targetPerson
                            socialPerson.groupAngle = math.atan(targetPerson.y - socialPerson.y, targetPerson.x - socialPerson.x)
                            socialPerson.groupOffset = math.random(-20, 20)
                            
                            socialPerson.groupTimer = socialPerson.groupTimer or 0
                            socialPerson.groupTimer = socialPerson.groupTimer + DeltaTime
                            if socialPerson.groupTimer > 5 then
                                socialPerson.group = nil
                                socialPerson.groupTimer = 0
                            end
                        end
                        
                        -- Bounce off each other
                        bounce(people[i], people[j])
                    end
                end
            end
        end
    end
    
    local function handleSocialBehavior()
        for i, person in ipairs(people) do
            if person.trait == "social" then
                -- Find the leader (if any)
                if person.group then
                    local leader = person.group
                    local followDistance = (person.size + leader.size) + person.followDistance
                    local minDistance = 10
                    local dx = leader.x - person.x
                    local dy = leader.y - person.y
                    local distance = math.sqrt(dx * dx + dy * dy)
                    
                    if distance > minDistance then
                        local targetX = leader.x - math.cos(leader.angle) * followDistance
                        local targetY = leader.y - math.sin(leader.angle) * followDistance
                        local targetAngle = math.atan(targetY - person.y, targetX - person.x)
                        local angleDifference = targetAngle - person.angle
                        
                        if angleDifference > math.pi then
                            angleDifference = angleDifference - 2 * math.pi
                        elseif angleDifference < -math.pi then
                            angleDifference = angleDifference + 2 * math.pi
                        end
                        
                        local maxAngleChange = math.rad(10)
                        angleDifference = math.max(-maxAngleChange, math.min(maxAngleChange, angleDifference))
                        person.angle = person.angle + angleDifference
                        
                        person.x = person.x + math.cos(person.angle) * person.speed * DeltaTime
                        person.y = person.y + math.sin(person.angle) * person.speed * DeltaTime
                    end
                    
                    person.groupTimer = person.groupTimer or 0
                    person.groupTimer = person.groupTimer + DeltaTime
                    if person.groupTimer > 5 then
                        person.group = nil
                        person.groupTimer = 0
                    end
                else
                    -- Find a new leader
                    for _, other in ipairs(people) do
                        if other ~= person and (other.trait ~= "social" or other.group ~= person) and checkCollision(person, other, person.size * 3) then
                            person.group = other
                            person.groupTimer = 0
                            break
                        end
                    end
                end
            elseif person.trait == "shy" then
                -- Shy people avoid collisions
                for j, other in ipairs(people) do
                    if i ~= j and checkCollision(person, other, person.size * 2) then
                        local dx = person.x - other.x
                        local dy = person.y - other.y
                        local distance = math.sqrt(dx * dx + dy * dy)
                        local nx = dx / distance
                        local ny = dy / distance
                        person.angle = math.atan(ny, nx)
                    end
                end
            end
        end
    end
    
    -- Main draw logic
    updateBackground()
    drawInstructions()
    handleSurpriseLines()
    updateTypeChangeTimer()
    updateAndDrawPeople()
    drawEmojis()
    handleCollisions()
    handleSocialBehavior()
end